#include "lights.h"
#include <GL/glut.h>


float zeroes[4] = { 0.0, 0.0, 0.0, 0.0 };

float ambient[4] = { 0.0, 0.0, 0.0, 1.0 };

// point light in the room
float light0[3][4] = { { 0.7, 0.7, 0.7, 1 },
                       { 0.5, 0.5, 0.5, 1 },
                       { 9, 9, 9, 1 } };

// spot light on robot's head
float light1[4][4] = { { 1, 1, 1, 1 },
                       { 1, 1, 1, 1 },
                       { 0, 0, 0.5, 1 },
                       { 0, 1, 0, 1 } };
float light1cutoff = 10;

// spot light in room
float light2[4][4] = { { 0.5, 0.5, 0.5, 1 },
                       { 0.5, 0.5, 0.5, 1 },
                       { -9.9, 9, 10.5, 1 },
                       { 0.5, 0.3, -1, 1 } };
float light2cutoff = 20;

// spot light in room
float light3[4][4] = { { 0.5, 0.5, 0.5, 1 },
                       { 0.5, 0.5, 0.5, 1 },
                       { 9, -9.9, 10.5, 1 },
                       { 0.3, 0.5, -1, 1 } };
float light3cutoff = 20;

// point light in the room
float light4[3][4] = { { 0.4, 0.4, 0.4, 1 },
                       { 0.2, 0.2, 0.2, 1 },
                       { -9, -9, 9, 1 } };



void make_lights ()
{
  glEnable(GL_LIGHTING);

  glLightfv(GL_LIGHT0, GL_AMBIENT,  zeroes);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0[0]);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light0[1]);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.1);
  glEnable(GL_LIGHT0);

  glLightfv(GL_LIGHT1, GL_AMBIENT,        zeroes);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,        light1[0]);
  glLightfv(GL_LIGHT1, GL_SPECULAR,       light1[1]);
  glLightf (GL_LIGHT1, GL_SPOT_CUTOFF,    light1cutoff);
  glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.05);

  glLightfv(GL_LIGHT2, GL_AMBIENT,        zeroes);
  glLightfv(GL_LIGHT2, GL_DIFFUSE,        light2[0]);
  glLightfv(GL_LIGHT2, GL_SPECULAR,       light2[1]);
  glLightf (GL_LIGHT2, GL_SPOT_CUTOFF,    light2cutoff);
  glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.1);
  glEnable(GL_LIGHT2);

  glLightfv(GL_LIGHT3, GL_AMBIENT,        zeroes);
  glLightfv(GL_LIGHT3, GL_DIFFUSE,        light3[0]);
  glLightfv(GL_LIGHT3, GL_SPECULAR,       light3[1]);
  glLightf (GL_LIGHT3, GL_SPOT_CUTOFF,    light3cutoff);
  glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 0.1);
  glEnable(GL_LIGHT3);

  glLightfv(GL_LIGHT4, GL_AMBIENT,  zeroes);
  glLightfv(GL_LIGHT4, GL_DIFFUSE,  light4[0]);
  glLightfv(GL_LIGHT4, GL_SPECULAR, light4[1]);
  glLightf(GL_LIGHT4, GL_LINEAR_ATTENUATION, 0.1);
  glEnable(GL_LIGHT4);

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

  place_lights();
}


void place_lights ()
{
  glLightfv(GL_LIGHT0, GL_POSITION,       light0[2]);
  glLightfv(GL_LIGHT2, GL_POSITION,       light2[2]);
  glLightfv(GL_LIGHT2, GL_SPOT_DIRECTION, light2[3]);
  glLightfv(GL_LIGHT3, GL_POSITION,       light3[2]);
  glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION, light3[3]);
  glLightfv(GL_LIGHT4, GL_POSITION,       light4[2]);
}


void set_ambient_light (float r, float g, float b)
{
  ambient[0] = r;
  ambient[1] = g;
  ambient[2] = b;
  ambient[3] = 1;

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
}


void place_spotlight ()
{
  glLightfv(GL_LIGHT1, GL_POSITION,       light1[2]);
  glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, light1[3]);
  glEnable(GL_LIGHT1);
}


void calc_shadow_matrix (float m[4][4], float p[4], float l[4])
{
  float dot = p[0]*l[0] + p[1]*l[1] + p[2]*l[2] + p[3]*l[3];

  m[0][0] = dot - l[0] * p[0];
  m[1][0] = 0.0 - l[0] * p[1];
  m[2][0] = 0.0 - l[0] * p[2];
  m[3][0] = 0.0 - l[0] * p[3];

  m[0][1] = 0.0 - l[1] * p[0];
  m[1][1] = dot - l[1] * p[1];
  m[2][1] = 0.0 - l[1] * p[2];
  m[3][1] = 0.0 - l[1] * p[3];

  m[0][2] = 0.0 - l[2] * p[0];
  m[1][2] = 0.0 - l[2] * p[1];
  m[2][2] = dot - l[2] * p[2];
  m[3][2] = 0.0 - l[2] * p[3];

  m[0][3] = 0.0 - l[3] * p[0];
  m[1][3] = 0.0 - l[3] * p[1];
  m[2][3] = 0.0 - l[3] * p[2];
  m[3][3] = dot - l[3] * p[3];
}
