#include "scene.h"

#include <GL/glut.h>
#include <iostream.h>
#include "robot.h"
#include "lights.h"
#include "main.h"


int drawFloor_list;
int drawWalls_list;


float floor_plane[4] = { 0, 0, 1, 0 };
float floor_shadow0[4][4];
float floor_shadow2[4][4];
float floor_shadow3[4][4];
float floor_shadow4[4][4];

GLuint texture_id;

bool drawing_shadows = false;
bool draw_mirror     = true;


float mesh_points[4][4][3] = { { { -1.5, -1.5,  4.0 },
                                 { -0.5, -1.5,  2.0 },
                                 {  0.5, -1.5, -1.0 },
                                 {  1.5, -1.5,  2.0 } },
                               { { -1.5,  0.5,  1.0 },
                                 { -0.5, -0.5,  3.0 },
                                 {  0.5, -0.5,  0.0 },
                                 {  1.5, -0.5, -1.0 } },
                               { { -1.5,  0.5,  4.0 },
                                 { -0.5,  0.5,  0.0 },
                                 {  0.5,  0.5,  3.0 },
                                 {  1.5,  0.5,  4.0 } },
                               { { -1.5,  1.5, -2.0 },
                                 { -0.5,  1.5, -2.0 },
                                 {  0.5,  1.5,  0.0 },
                                 {  1.5,  1.5, -1.0 } } };


void initSceneDisplayLists ()
{
  drawFloor_list = glGenLists(FLOOR_D); 
  glNewList(drawFloor_list, GL_COMPILE);
  drawFloor();
  glEndList();

  drawWalls_list = glGenLists(WALLS_D); 
  glNewList(drawWalls_list, GL_COMPILE);
  drawWalls();
  glEndList();
}


void drawEverything ()
{
  glMatrixMode(GL_MODELVIEW);

  drawFloorReflection();
  if (draw_mirror)
      drawMirrorReflection();

  place_lights();

  set_spotlight = true;
  drawRobot();
  drawRoom();
  drawShadows();
}


void drawFloorReflection ()
{
  glDisable(GL_DEPTH_TEST);
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 0x1, 0x1);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);

  glCallList(drawFloor_list);

  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glEnable(GL_DEPTH_TEST);

  glStencilFunc(GL_EQUAL, 0x1, 0x1);
  glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

  glPushMatrix();
    glScalef(1, 1, -1);
    place_lights();

    glCullFace(GL_FRONT);

    set_spotlight = false;
    glCallList(drawWalls_list);
    drawSolids();
    drawRobot();

    glCullFace(GL_BACK);

  glPopMatrix();

  glDisable(GL_STENCIL_TEST);
}


void drawMirrorReflection ()
{
  glDisable(GL_DEPTH_TEST);
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 2, 0x2);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);

  drawMirror();

  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glEnable(GL_DEPTH_TEST);

  glStencilFunc(GL_EQUAL, 2, 0x2);
  glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

  glClipPlane(GL_CLIP_PLANE0, 
              (GLdouble[4]){ -1.0, 0.0, 0.0, -10.0 });
  glEnable(GL_CLIP_PLANE0);

  glPushMatrix();
    glScalef(-1, 1, 1);
    glTranslatef(20, 0, 0);
    place_lights();

    glCullFace(GL_FRONT);

    set_spotlight = false;
    glCallList(drawFloor_list);
    glCallList(drawWalls_list);
    drawRobot();
    drawSolids();

    glCullFace(GL_BACK);

  glPopMatrix();

  glDisable(GL_CLIP_PLANE0);
  glDisable(GL_STENCIL_TEST);

  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  if (draw_mirror)
      drawMirror();
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}


void drawRoom ()
{
  glFrontFace(GL_CW);
  glCallList(drawWalls_list);
  glCallList(drawFloor_list);
  glFrontFace(GL_CCW);

  glCallList(drawWalls_list);

  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 0x4, 0x4);
  glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glCallList(drawFloor_list);
  glDisable(GL_BLEND);

  glDisable(GL_STENCIL_TEST);

  drawSolids();
}


void drawMirror ()
{
  glPushMatrix();
    glMaterialfv(GL_FRONT, GL_AMBIENT,
                 (float[4]){ 0.1, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT, GL_DIFFUSE,
                 (float[4]){ 0.1, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT, GL_SPECULAR,
                 (float[4]){ 0, 0, 0, 1 });

    glBegin(GL_POLYGON);
    glNormal3f(0, 1, 0);
    glVertex3f(-10, 2, 1);
    glVertex3f(-10, 6, 1);
    glVertex3f(-10, 6, 8);
    glVertex3f(-10, 2, 8);
    glEnd();

  glPopMatrix();
}


void drawSolids ()
{
  glPushMatrix();
  glTranslatef(-15, -10, 0);

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.1, 0.7, 0.2, 1 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 0.1, 0.7, 0.2, 1 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 0.1, 0.3, 0.1, 1 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.1 });

    glutSolidCone(2, 10, 15, 10);

  glPopMatrix();


  glPushMatrix();
    glCullFace(GL_FRONT);
    if (draw_mirror)
        drawMirror();
    glCullFace(GL_BACK);
  glPopMatrix();

  glPushMatrix();
    glTranslatef(0, 10, 4);

    glDisable(GL_CULL_FACE);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,
                 (float[4]){ 0.7, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,
                 (float[4]){ 0.7, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,
                 (float[4]){ 0.5, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS,
                 (float[1]){ 0.5 });

    glEvalMesh2(GL_FILL, 0, 20, 0, 20);

    glEnable(GL_CULL_FACE);

  glPopMatrix();

  glPushMatrix();
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texture_id);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,
                 (float[4]){ 0.7, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,
                 (float[4]){ 0.7, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,
                 (float[4]){ 0.5, 0.1, 0.1, 1 });
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS,
                 (float[1]){ 0.5 });

    glBegin(GL_POLYGON);
    glTexCoord2f(0.0,0.0);
    glVertex3f(10.0,8.0,2.0);
    glTexCoord2f(0.0,1.0);
    glVertex3f(10.0,8.0,4.0);
    glTexCoord2f(1.0,1.0);
    glVertex3f(8.0, 10.0, 4.0);
    glTexCoord2f(1.0,0.0);
    glVertex3f(8.0, 10.0, 2.0);
    glEnd();

    glBegin(GL_POLYGON);
    glTexCoord2f(0.0,0.0);
    glVertex3f(10.0,8.0,2.0);
    glTexCoord2f(1.0,0.0);
    glVertex3f(8.0, 10.0, 2.0);
    glTexCoord2f(1.0,1.0);
    glVertex3f(8.0, 10.0, 4.0);
    glTexCoord2f(0.0,1.0);
    glVertex3f(10.0,8.0,4.0);
    glEnd();

    glDisable(GL_TEXTURE_2D);
  glPopMatrix();

  glPushMatrix();
    glTranslatef(5, -5, 3.001);
    glRotatef(45, 0, 0, 1);
    glScalef(2, 1, 6);

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.4, 0.2, 0.7, 0.7 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 0.4, 0.2, 0.7, 0.7 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 0.2, 0.1, 0.3, 0.7 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.5 });

    if (!drawing_shadows)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }

    glutSolidCube(1);

    if (!drawing_shadows)
        glDisable(GL_BLEND);

  glPopMatrix();
}


void drawWalls ()
{
  glPushMatrix();
  glTranslatef(20, 0, 0);
  glRotatef(-90, 0, 1, 0);

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.5, 0.6, 0.5, 1 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 0.6, 0.7, 0.6, 1 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 0.4, 0.6, 0.4, 1 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.9 });

  drawTiledRect(0, -20, 10, 20, 20, 40, NULL);

  glPopMatrix();
  glPushMatrix();
  glTranslatef(0, 20, 0);
  glRotatef(90, 0, 0, 1);
  glRotatef(-90, 0, 1, 0);

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.5, 0.5, 0.6, 1 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 0.6, 0.6, 0.7, 1 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 0.4, 0.4, 0.6, 1 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.9 });

  drawTiledRect(0, -20, 10, 20, 20, 40, NULL);

  glPopMatrix();
}


void drawShadows ()
{
  set_spotlight = false;
  drawing_shadows = true;

  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_LESS, 0, 0x4);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);

  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(-1.0, -1.0);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_LIGHTING);
  glColor4f(0.0,0.0,0.0,0.5);

  glPushMatrix();
    glMultMatrixf((GLfloat*) floor_shadow0);
    drawRobot();
    drawSolids();
    if (draw_mirror)
        drawMirror();
  glPopMatrix();
  glPushMatrix();
    glMultMatrixf((GLfloat*) floor_shadow4);
    drawRobot();
    drawSolids();
    if (draw_mirror)
        drawMirror();
  glPopMatrix();

  glEnable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glDisable(GL_STENCIL_TEST);
  glDisable(GL_POLYGON_OFFSET_FILL);

  drawing_shadows = false;
}


bool floor_pattern_a (int x, int y)
{
  return (((x + y) % 2) == 0);
}


bool floor_pattern_b (int x, int y)
{
  return (((x + y) % 2) == 1);
}


void drawFloor ()
{
  glPushMatrix();

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.5, 0.5, 0.5, 0.2 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 1, 1, 1, 0.2 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 1, 1, 1, 0.2 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.9 });

  drawTiledRect(-20, -20, 20, 20, 40, 40, floor_pattern_a);

  glMaterialfv(GL_FRONT, GL_AMBIENT,
               (float[4]){ 0.5, 0.5, 0.5, 0.5 });
  glMaterialfv(GL_FRONT, GL_DIFFUSE,
               (float[4]){ 0.6, 0.6, 0.6, 0.5 });
  glMaterialfv(GL_FRONT, GL_SPECULAR,
               (float[4]){ 1, 1, 1, 0.5 });
  glMaterialfv(GL_FRONT, GL_SHININESS,
               (float[1]){ 0.9 });

  drawTiledRect(-20, -20, 20, 20, 40, 40, floor_pattern_b);

  glPopMatrix();
}


void drawTiledRect (float x1, float y1, float x2, float y2,
                    int divx, int divy, bool (*pred)(int,int))
{
  float x  = x1;
  float y  = y1;
  float dx = (x2 - x1) / divx;
  float dy = (y2 - y1) / divy;

  for (int i = 0; i < divx; i++)
  {
      for (int j = 0; j < divy; j++)
      {
          if ((pred == NULL) || pred(i,j))
          {
              glBegin(GL_POLYGON);
              glNormal3f(0,0,1);
              glVertex3f(x,y,0);
              glVertex3f(x+dx,y,0);
              glVertex3f(x+dx,y+dy,0);
              glVertex3f(x,y+dy,0);
              glEnd();
          }

          y += dy;
      }
      x += dx;
      y = y1;
  }
}



#define TEXTURE_WIDTH	16
#define TEXTURE_HEIGHT	16

static char *texture_data[] = {
  "....xxxx........", 
  "..xxxxxxxx......", 
  ".xxxxxxxxxx.....", 
  ".xxx....xxx.....",
  "xxx......xxx....",
  "xxx......xxx....", 
  "xxx......xxx....",    
  "xxx......xxx....", 
  ".xxx....xxx.....",
  ".xxxxxxxxxx.....",
  "..xxxxxxxx......",
  "....xxxx........",
  "................",
  "................",
  "................",
  "................",
};

void initTexture ()
{
  GLubyte  texture_pixels[TEXTURE_WIDTH][TEXTURE_HEIGHT][3];
  GLubyte* pixel = (GLubyte*) texture_pixels;

  for (int t = 0; t < TEXTURE_WIDTH; t++)
      for (int s = 0; s < TEXTURE_HEIGHT; s++)
      {
          if (texture_data[t][s] == 'x')
          {
              pixel[0] = 0x1f;
              pixel[1] = 0x8f;
              pixel[2] = 0x1f;
          }
          else
          {
              pixel[0] = 0xaa;
              pixel[1] = 0xaa;
              pixel[2] = 0xaa;
          }
          pixel += 3;
      }

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &texture_id);
  glBindTexture(GL_TEXTURE_2D, texture_id);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, TEXTURE_WIDTH, TEXTURE_HEIGHT, 
               0, GL_RGB, GL_UNSIGNED_BYTE, texture_pixels);
}
